// alphabet.h -- The -*- C++ -*- alphabet class interface.
// Copyright (C) 1997 Sergio Talens-Oliag <sto@uv.es>
// This file is part of the C++ Alphabetic Data Management Library.

#ifndef _ALPHABET_H_
#define _ALPHABET_H_

#include <iostream.h> 
#include <iomanip.h> 
#include <map> 
#include <set> 
#include <vector> 

template <class T>
class alphabet {
public:
  // typedefs
  typedef vector<T>::size_type size_type;
  typedef less<T>              lessT;

private:

  vector<T>                    svect; // map index to symbols
  map<T, size_type, lessT>     smap;  // map symbols to indexes

public:

  // constructors / destructor

  // default constructor

  // The default generic constuctor is useless, define a new one or
  // call the algorithms with an initialized alphabet. Has to be here
  // to allow the declaration of the algorithms without parameters.

  alphabet () : svect(), smap() {};

  // constructor w./ initialization
  alphabet (const vector<T> &symbols) {
 
    set<T, lessT>            sset;
    set<T, lessT>::iterator  si;
    size_type                i;

    // insert symbols in a set to remove duplicate entries
    sset.insert(symbols.begin(), symbols.end());  
    
    // reserve the space for the symbols
    svect.reserve(sset.size());
    
    for (i=0, si=sset.begin(); i < sset.size(); i++, si++) {
      svect.push_back(*si);
      smap[*si] = i;
    }

  };

  // Copy constructor
  alphabet (const alphabet<T> &a) : svect(a.svect), smap(a.smap) {};

  ~alphabet() {};

  // Copy operator -- not really needed
  alphabet<T> & operator= (const alphabet<T> &a) {
    svect = a.svect;
    smap  = a.smap;
    return *this;
  };

  // access methods
  // should be a const method, but with gcc fails because of smap[s]
  size_type index(const T& s) { 
    return (belongs(s) ? smap[s] : size()); 
  }; // which is the index of the symbol ? (size() if not in the alphabet)

  const T&  value(const size_type &i) const {
    return svect[i];
  }; // which is the symbol pointed by i ?

  bool belongs(const T& s) const { 
    return (smap.find(s) != smap.end());
  }; // is the symbol on the alphabet?

  size_type size() const { 
    return svect.size(); 
  }; // size of the alphabet

  bool empty() const { 
    return svect.empty();
  }; // Is the alphabet empty

  // STATIC METHODS
  static size_type max_size(); // HAS TO BE DEFINED FOR EACH TYPE T
};

// Specialization for type CHAR

/*
  NOTE:

  The ANSI C++ STANDARD DRAFT when talking about integral conversions
  says (4.7, paragraph 2):

  "If  the destination type is unsigned, the resulting value is the least
   unsigned integer congruent to the source integer (modulo 2n where n is
   the  number of bits used to represent the unsigned type).  [Note: In a
   two's complement representation, this  conversion  is  conceptual  and
   there is no change in the bit pattern (if there is no truncation).  ]  "

  But G++ does the following: 

    char c = '';

    unsigned char (c) == 241
    unsigned int  (c) == 4294967281

  Because of that, indexes are converted to unsigned char before
  being converted to size_type (usually unsigned int).

  The same discussion is applicable for signed char, but we're not
  providing versions of alphabet for signed nor unsigned char because
  string does not provide signed or unsigned char strings.

*/

class alphabet<char> {
public:

  // typedefs
  typedef vector<char>::size_type  size_type;
  typedef less<char>               lesschar;

private:
  vector<char>                     svect;  // map index to symbols
  vector<size_type>                smap;   // map symbols to indexes

public:

  // constructors / destructor

  // Default constructor, this one defines the identity alphabet
  alphabet () : svect(max_size()), smap(max_size()) {
    for (unsigned int i=0; i < max_size(); i++) {
      smap[i]=i;
      svect[i]=i;
    }
  };

  alphabet (const vector<char> &symbols) {

    set<char, lesschar>           sset;
    set<char, lesschar>::iterator si;
    size_type                     i;

    // insert symbols in a set to remove duplicate entries
    sset.insert(symbols.begin(), symbols.end());  
    
    // reserve the space for the symbols
    svect.reserve(sset.size());
    
    // set all smap values to sset.size()
    smap.insert(smap.begin(), max_size(), sset.size());

    for (i=0, si=sset.begin(); i < sset.size(); i++, si++) {
      svect.push_back(*si);
      smap [ (unsigned char)(*si) ] = i;
    }
  };

  alphabet (const alphabet<char> &a) : svect(a.svect), smap(a.smap) {};
  
  ~alphabet() {};

  // Copy operator -- not really needed
  alphabet<char> & operator= (const alphabet<char> &a) {
    svect = a.svect;
    smap  = a.smap;
    return *this;
  };

  // access methods
  size_type index(const char& s) const { 
    return smap[ (unsigned char) s ];         // See NOTE
  }; // which is the index of the symbol ? (size() if not in the alphabet)

  const char& value(size_type i) const { 
    return svect[i]; 
  }; // which is the symbol pointed by i ?

  bool belongs(const char& s) const { 
    return index(s) < svect.size();    
  }; // is the symbol on the alphabet ?

  size_type size() const {
    return svect.size(); 
  }; // size of the alphabet

  bool empty() const { 
    return svect.empty();
  }; // is the alphabet empty ?

  // STATIC METHODS
  static size_type max_size() { 
    return 256; 
  };  // maximum size of the alphabet
};

// ALPHABET OUTPUT
template <class T>
ostream & operator<< (ostream &o, alphabet<T> &a) {
  alphabet<T>::size_type i;

  o << "ALPHABET SIZE = " << a.size() << "\n";
  o << "VALUES        = ";
  for (i=0; i < a.size(); i++)
    o << setw(5) << a.value(i);
  o << "\n";

  return o;
};

#endif
