// atoken.h -- The -*- C++ -*- alphabetic token class interface.
// Copyright (C) 1997 Sergio Talens-Oliag <sto@uv.es>
// This file is part of the C++ Alphabetic Data Management Library.

#ifndef _ATOKEN_H_
#define _ATOKEN_H_

/**

  DESCRIPTION

  Class to represent a token. A set of adaption functions is defined
  to get the begining and end of a token in a string and to verify if
  a substring is a token.

  We define three classes, the base class (atoken) has the common
  functions and two derived versions are defined to allow two
  different token definitions, by limits or by content:

    adtoken (delimited token): 

      The token is defined by it's limits, it's built from two symbol
      lists, the pretoken and posttoken sets.

    actoken (content token):

      The token is defined by it's contents, it's built from a symbol
      list.

**/

#include "astring.h"

#ifdef _CHAR_ASTRING_
# include <vector>
#else
# include <set>
#endif

// ==========================================
// ATOKEN (Alphabetic Token): Class Interface
// ==========================================

class atoken {

public:

  typedef astring::iterator       iterator;
  typedef astring::const_iterator const_iterator;
  typedef astring::size_type      size_type;
  typedef astring::value_type     value_type;

protected:

  const char *                    name;

public:

  atoken(const char *tn) : name(tn) {};
  ~atoken(){};

  virtual
  bool is_predelim  (const value_type &d) const = 0;

  virtual
  bool is_postdelim (const value_type &d) const = 0;

  virtual
  bool is_delim     (const value_type &d) const = 0;

  inline
  bool is_content   (const value_type &d) const;

  // Determine if the text between tb & te is a token (don't look at
  // it's limits)
  bool is(const astring &t, const_iterator tb, const_iterator te) const;

  inline
  bool is(const astring &t, const size_type tb, const size_type te) const;

  // Return the beginning of a token in t starting BEFORE ti
  // If ti is not pointing to the text return ti
  iterator begin(const astring &t, const_iterator ti) const;

  inline
  astring::size_type begin(const astring &t, const size_type ti) const;

  // Return the end of a token in t starting IN ti
  // If ti is not pointing to the text return ti
  astring::iterator end(const astring &t, const_iterator ti) const;

  inline
  astring::size_type end(const astring &t, const size_type te) const;

  // Get a token in t starting IN or BEFORE tb and return it's end
  inline
  astring::iterator get(const astring &t, iterator &tb) const;
  
  inline
  astring::size_type get(const astring &t, size_type &tb) const;
};

// ====================================================
// ACTOKEN (Alphabetic Contents Token): Class Interface
// ====================================================

class actoken : public atoken {

public:

  typedef astring::value_type       value_type;

protected:

#ifdef _CHAR_ASTRING_
  vector<bool>                      nvs;
#else
  set<value_type, less<value_type>> vss;
#endif

public:

  inline
  actoken(const char *tn, const astring &vs);

  inline
  ~actoken();

  inline
  bool is_predelim (const value_type &d) const;

  inline
  bool is_postdelim (const value_type &d) const;

  inline
  bool is_delim (const value_type &d) const;
};

// =====================================================
// ADTOKEN (Alphabetic Delimited Token): Class Interface
// =====================================================

class adtoken : public atoken {

protected:

#ifdef _CHAR_ASTRING_
  vector<bool>                      pre;
  vector<bool>                      post;
#else
  set<value_type, less<value_type>> pre;
  set<value_type, less<value_type>> post;
#endif

public:

  inline
  adtoken(const char *tn, const astring &pr, const astring &po);
 
  inline
  ~adtoken();

  inline
  bool is_predelim  (const value_type &d) const;
  inline
  bool is_postdelim (const value_type &d) const;
  inline
  bool is_delim     (const value_type &d) const;
};

// ==============
// TOKEN ADAPTORS
// ==============

inline
astring::iterator  begin(atoken *t, const astring &s,
                         const astring::iterator ts) { 
  return t->begin(s, ts); 
};

inline
astring::size_type begin(atoken *t, const astring &s,
                         const astring::size_type ts) { 
  return t->begin(s, ts); 
};

inline
astring::iterator  end(atoken *t, const astring &s,
		       const astring::iterator ts) { 
  return t->begin(s, ts); 
};

inline
astring::size_type end(atoken *t, const astring &s,
		       const astring::size_type ts) { 
  return t->begin(s, ts); 
};

inline 
bool is(atoken *t, const astring &s, const astring::iterator tb,
	const astring::iterator te) { 
  return t->is(s, tb, te); 
};

inline 
bool is(atoken *t, const astring &s, const astring::size_type tb,
	const astring::size_type te ) { 
  return t->is(s, tb, te); 
};

// =========================================
// ATOKEN (Alphabetic Token): Inline Methods
// =========================================

inline
bool atoken::is_content(const value_type &d) const {
  return !(is_delim(d));
};  

inline
bool atoken::is(const astring &t, const size_type tb,
		const size_type te) const {
  return is(t, t.begin() + tb, t.begin() + te);
};

// Get token begin
inline
astring::size_type atoken::begin(const astring &t, const size_type ti) const {
  return begin(t, t.begin() + ti) - t.begin();
};

// Get token end
inline
astring::size_type atoken::end(const astring &t, const size_type te) const {
  return end(t, t.begin() + te) - t.begin();
};

// Get a token in t starting IN or BEFORE tb and return it's end
inline
astring::iterator atoken::get(const astring &t, iterator &tb) const {
  iterator te = iterator(tb + 1); // the token can end just after tb
  tb = begin(t, tb);
  return end(t, te);
};
  
inline
astring::size_type atoken::get(const astring &t, size_type &tb) const {
  size_type te = tb + 1; // the token can end just after tb
  tb = begin(t, tb);
  return end(t, te);
};


// ===================================================
// ACTOKEN (Alphabetic Contents Token): Inline Methods
// ===================================================

#ifdef _CHAR_ASTRING_

inline
actoken::actoken(const char *tn, const astring &vs)
  : atoken(tn), nvs(256, true) {
    astring::iterator i;
  
    for (i=(astring::iterator)vs.begin(); i < vs.end(); i++)
      nvs[(unsigned char)(*i)] = false;
};

inline
bool actoken::is_delim (const value_type &d) const {
  return nvs[(unsigned char)(d)];
};

#else

inline
actoken::actoken(const char *tn, const astring &vs)
  : atoken(tn), con(vs.begin(), vs.end()) {};

inline
bool actoken::is_delim (const value_type &d) const {
  return (vss.find(d) == con.end());
};

#endif

inline
actoken::~actoken() {};

inline
bool actoken::is_predelim  (const value_type &d) const {
  return is_delim(d);
};

inline
bool actoken::is_postdelim (const value_type &d) const {
  return is_delim(d);
};

// ====================================================
// ADTOKEN (Alphabetic Delimited Token): Inline Methods
// ====================================================

#ifdef _CHAR_ASTRING_

inline
adtoken::adtoken(const char *tn, const astring &pr, const astring &po) 
  : atoken(tn), pre(256, false), post(256, false) {

    astring::iterator i;

    for (i=(astring::iterator)pr.begin(); i < pr.end(); i++)
      pre[(unsigned char)(*i)] = true;

    for (i=(astring::iterator)po.begin(); i < po.end(); i++)
      post[(unsigned char)(*i)] = true;

};

inline
bool adtoken::is_predelim  (const value_type &d) const {
  return pre[(unsigned char)(d)];
};

inline
bool adtoken::is_postdelim (const value_type &d) const {
  return post[(unsigned char)(d)];
};

#else

adtoken::adtoken(const char *tn, const astring &pr, const astring &po) 
  : atoken(tn), pre(pr.begin(), pr.end()), post(po.begin(), po.end()) {};

inline
bool adtoken::is_predelim  (const value_type &d) const {
  return (pre.find(d) != pre.end());
};

inline
bool adtoken::is_postdelim (const value_type &d) const {
  return (post.find(d) != post.end());
};

#endif

inline
adtoken::~adtoken() {};

inline
bool adtoken::is_delim(const value_type &d) const {
  return (is_predelim(d) || is_postdelim(d));
};

#endif
