// bmhskm.cpp -- The -*- C++ -*- Boyer Moore Horspool skm prep & matcher. 
// Copyright (C) 1997 Sergio Talens-Oliag <sto@uv.es>
// This file is part of the C++ Alphabetic Data Management Library.

#ifdef BMHSKM_CC

// Boyer Moore Horspool preprocessor
void bmh_skm::PREPROCESSOR () {

  // internal variables
  skm_patt::size_type i;

  // INIT SKIP TABLE

  // clear skip vector
  skip.erase(skip.begin(), skip.end());            // erase previous values
  skip.reserve(alpha_size);                        // inform of space needed

  // allocate vector with assignment (skip[i] = pat_size)
  skip.insert(skip.begin(), alpha_size, pat_size); 

  // compute skip & pat_last
  pat_last = pat_beg;

#ifdef BMHSKM_RUN
  report_step(INIT_PREPROCESSOR); // Step 0
  if (stop) return;

  report_step(PREPROCESSOR_STEP, alpha_size);
  // Inicialization of skip -> no psteps, but they are there
  if (stop) return;
#endif

  for (i= pat_size - 1; i > 0; i--) {
    skip [ aindex(*pat_last) ] = i;
#ifdef BMHSKM_RUN
    report_step(PREPROCESSOR_STEP); // skip updates
    if (stop) return;
#endif
    pat_last++;
  }

  /*
     Now we have:

      skip[pat[i-1]] == pat_size - i   
      pat_last       == pat_end - 1
  */
};

/**
  
  NOTE:

  skip[*pat_last] is not assigned a value, so it's never 0 (it'll be
  'pat_size' or less if *pat_last is found previously in the pattern).
  
  This is so to advance the machine from the last pattern symbol after
  a partial match, instead of doing it from the current text mismatch
  position (Horspool improvement).
  
  Why? This avoids some comparisions of symbols already seen and
  allows the recovery of the machine after a match.
  
  With the Simplified Boyer Moore aproach (using only the skip table),
  when there's a mismatch after a partial match we restart from 'skip
  [*pmti]' (where 'pmti' is the position of the mismatched symbol in
  the text), but we've already seen the symbols between 'pmti + 1' and
  'ti' ('ti' is the position that started the partial match, '*ti ==
  *pat_last'), so we can continue from 'skip[*ti]':

      text     :   g  o  o  d  _  m  o  r  n  i  n  g 

      pattern  :   h  o  o  d

      bmh_skip :   3  1  1  4

      sbm_skip :   3  1  1  0

      [g][*ood*]_morning -> SBM(skip['g'] == 4) -> good[_]morning
      [g][*ood*]_morning -> BMH(skip['d'] == 4) -> good_mo[r]ning
    
  The same is true when recovering from a match, in the SBM model we
  have to continue from 'ti + 1' to avoid loosing overlapping matches:

      text     :   t  o  m  a  t  o  m  a  t  o

      pattern  :   t  o  m  a  t  o

      bmh_skip :   1  4  3  2  1  4

      sbm_skip :   1  0  3  2  1  0
 
      [*tomato*]mato -> SBM                  -> tomato[m]ato
      [*tomato*]mato -> BMH (skip['o'] == 4) -> tomatomat[o]

  If *pat_last is not repeated in the text we'll advance as with any
  symbol NOT IN the pattern (pat_size - 1), if it's repeated we'll
  advance as with any other symbol IN the pattern.

**/

//  Boyer Moore Horspool matcher
astring::size_type bmh_skm::MATCHER (){

  // LOCAL OPTIMIZATIONS

  // USE pl INSTEAD OF *pat_last 
  register skm_patt::value_type pl = *pat_last;

#ifndef BMHSKM_RUN
  // Redefine the text iterator declaring it in register
  register astring::iterator   ti = this->ti;
#endif

  ti += pat_size - 1; // advance the text iterator to be aligined with pat_last

#ifdef BMHSKM_RUN
  report_step(INIT_MATCHER); // Step 0, get first text positions
  if (stop) return txt_size;
#endif

  while (ti < txt_end) { 

    // Do partial match if current text symbol == last pattern symbol
    if (*ti==pl) {

      pmti = ti;
      pmpi = pat_last;

#ifdef BMHSKM_RUN
      report_step(PARTIAL_MATCH); // current text symbol = current patt symbol
      if (stop) return txt_size;
#endif

      // Use do-while instead of while to allow patterns of size one.
      do {

	// If pattern matched apply match function
	if (pmpi == pat_beg) {

	  // Internal variables
	  astring::size_type midx; // match index 

	  midx = pmti - txt_beg;

#ifdef BMHSKM_RUN
	  report_step(FULL_MATCH);
	  if (stop) return txt_size;
#endif

    	  if (report_match(midx)) {
	    return (midx);                // ... return the match
	  } else {
#ifdef BMHSKM_RUN
	    report_step(MATCH_RECOVERY);
	    if (stop) return txt_size;
#endif
	    break;                        // ... and continue searching
	  } // end if

	} // end if

	pmpi--; pmti--;

#ifdef BMHSKM_RUN
	report_step(PARTIAL_MATCH_STEP); // advance pattern & text
	if (stop) return txt_size;
#endif

      } while (*pmpi==*pmti);

    } // end if

    //    ti += skip[ aindex(*ti) ];
    ti += skip[ aindex(*ti) ];

#ifdef BMHSKM_RUN
    report_step(MATCHER_STEP); // advance text 
    if (stop) return txt_size;
#endif

  } // end while

  return (txt_size);
};

#endif

