// bmskm.cpp -- The -*- C++ -*- Boyer Moore skm preprocessor & matcher.
// Copyright (C) 1997 Sergio Talens-Oliag <sto@uv.es>
// This file is part of the C++ Alphabetic Data Management Library.

#ifdef BMSKM_CC

// Boyer Moore preprocessor
void bm_skm::PREPROCESSOR () {

  // INTERNAL VARIABLES
  vector<skm_patt::size_type>::iterator  shi;  
  skm_patt::difference_type              i, j, k;

  // INIT SKIP TABLE

  // clear skip vector
  skip.erase(skip.begin(), skip.end());            // erase values
  skip.reserve(alpha_size);                        // inform of space needed

  // allocate skip vector with assignment (skip[i] = pat_size)
  skip.insert(skip.begin(), alpha_size, pat_size);

  // INIT SHIFT TABLE

  // clear shift vector 
  shift.erase(shift.begin(), shift.end());         // erase values
  shift.reserve(pat_size);                         // inform of space needed

  // allocate shift vector with assignement (shift[i] = pat_size)
  shift.insert(shift.begin(), pat_size, pat_size);

  // SET UP SKIP (& SHIFT)
  pat_last = pat_beg;       // skip
  shi      = shift.begin(); // shift

#ifdef BMSKM_RUN
  report_step(INIT_PREPROCESSOR); // Step 0
  if (stop) return;

  report_step(PREPROCESSOR_STEP, alpha_size);
  // Inicialization of skip -> no psteps, but they are there
  if (stop) return;
#endif

  for (i = pat_size - 1; i >= 0; i--) {

    skip[ aindex(*pat_last) ] = i;  // skip
    *shi += i;                      // shift
 
#ifdef BMSKM_RUN
  report_step(PREPROCESSOR_STEP); // shift & skip updates
  if (stop) return;
#endif

    pat_last++;                     // skip
    shi++;                          // shift

  }

  // Now we have:
  //
  //   skip[pat[i-1]] == pat_size - i
  //
  // And:
  //
  //   shift[i-1] = 2 * pat_size - i

  k = pat_size + 1;  

  // f vector 
  f.reserve(pat_size);                         // inform of space needed
  // allocate f vector with assignement (f[i] = 0)
  f.insert(f.begin(), pat_size, 0);
 
  for (j=pat_size; j > 0; j--) {

    f[j-1] = k;

#ifdef BMSKM_RUN
    report_step(PREPROCESSOR_STEP); // f updates
    if (stop) return;
#endif
    
    while (k <= pat_size && (*pat)[j-1] != (*pat)[k-1]) {

      shift[k-1] = min ( shift[k-1], pat_size - j);      
      k = f[k-1];

#ifdef BMSKM_RUN
      report_step(PREPROCESSOR_STEP); // shift updates
      if (stop) return;
#endif

    } // end while

    k--;

  } // end for

  for (i=1; i<=k; i++) {
    shift[i-1] = min( shift[i-1], pat_size + k - i );
#ifdef BMSKM_RUN
    report_step(PREPROCESSOR_STEP); // shift updates
    if (stop) return;
#endif
  } // end for

  // Rytter correction as outlined by Mehlhorn (*)
 
  j = f[k-1];

  while(k < pat_size){

    while (k <= j) {

      shift[k-1] = min( shift[k -1], pat_size + j - k);
      k++;

#ifdef BMSKM_RUN
      report_step(PREPROCESSOR_STEP); // shift updates
      if (stop) return;
#endif

    }

    j = f[j-1];
  }

  // SET PAT LAST
  pat_last = pat_end - 1;

  // Erase f vector
  f.erase(f.begin(), f.end());         // erase values  
};

// (*) SEE:
//
// - AHO, Alfred V., "Algorithms for finding patterns in strings", in
// Leeuwen, J. ed., "Handbook of Theoretical Computer Science",
// Elsevier, New York, pp. 257-300 (1990).
//
// - Melhorn, K., "Data Structures and Algorithms 1: Sorting and
// Searching", Springer, Berlin (1984).

// Boyer Moore matcher
astring::size_type bm_skm::MATCHER (){

  // LOCAL OPTIMIZATIONS

  // USE pl INSTEAD OF *pat_last 
  register skm_patt::value_type pl = *pat_last;

  // USE shift_pl INSTEAD OF shift[pat_last - pat_beg];
  // register // Gives a warning with STL's max ??? 
  skm_patt::size_type           shift_pl = shift[pat_last - pat_beg];
  
#ifndef BMSKM_RUN
  // Redefine the internal pointers declaring them in registers
  register astring::iterator   ti = this->ti;
#endif

  ti += pat_size - 1; // advance the text iterator to be aligined with pat_last

#ifdef BMSKM_RUN
  report_step(INIT_MATCHER); // Step 0, get first text positions
  if (stop) return txt_size;
#endif

  while (ti < txt_end) { 

    // Do partial match if current text symbol == last pattern symbol
    if (*ti==pl) {

      pmpi = pat_last;

#ifdef BMSKM_RUN
      report_step(PARTIAL_MATCH); // current text symbol = current patt symbol
      if (stop) return txt_size;
#endif

      // Use do-while instead of while to allow patterns of size one.
      do {

	// If pattern matched apply match function
	if (pmpi == pat_beg) {

	  // Internal variables
	  astring::size_type midx; // match index 

	  midx = ti - txt_beg;

#ifdef BMSKM_RUN
	  report_step(FULL_MATCH);
	  if (stop) return txt_size;
#endif

    	  if (report_match(midx)) {
	    return (midx);                // ... return the match
	  } else {
#ifdef BMSKM_RUN
	    report_step(MATCH_RECOVERY);
	    if (stop) return txt_size;
#endif
	    break;                        // continue searching

	  } // endif

	} // endif

	pmpi--; ti--;

#ifdef BMSKM_RUN
	report_step(PARTIAL_MATCH_STEP); // text symbol = patt symbol
	if (stop) return txt_size;
#endif

      } while (*pmpi==*ti);

      ti += max(skip[ aindex(*ti) ], shift[pmpi - pat_beg]);
    
    } else {

      ti += max(skip[ aindex(*ti) ], shift_pl);

    } // end if

#ifdef BMSKM_RUN
    report_step(MATCHER_STEP); // advance text 
    if (stop) return txt_size;
#endif

  } // end while

  return (txt_size);
};

#endif
