// krskm.h -- The -*- C++ -*- Karp Rabin skm interface.
// Copyright (C) 1997 Sergio Talens-Oliag <sto@uv.es>
// This file is part of the C++ Alphabetic Data Management Library.

#ifndef _KRSKM_H_
#define _KRSKM_H_

#include "skm.h"
#include <limits.h> // UINT_MAX
#include <vector>   // skip table

// =====================================
// KR SKM (Karp Rabin) : Auxiliary table
// =====================================

const unsigned int PRIMES[] = {
  0, 
  1073741789, 536870909, 268435399, 134217689, 67108859, 
  33554393,   16777213,  8388593,   4194301,   2097143,
  1048573,    524287,    262139,    131071,    65521,
  32749,      16381,     8191,      4093,      2039, 
  1021,         509,     251,       127,       61,
  31,            13,     7,         3,         2,
  1
};

/**

  'alpha_size' has to be <= B (the hashing function represents each
  string as a number in base B, where B is the number of possible
  characters ( == alpha_size) ).
  
  B and Q are computed in the preprocessor using the alpha_size and
  the table of PRIMES.
  
  The methods use the aindex() to avoid overflow in hpat and htxt.
  
  We choose B to be a power of 2 to substitute the multiplications
  by shifts. [ A * B == A << BS if B = pow(2, BS) ].
  
**/

// =====================================
// KR SKM (Karp Rabin) : Class Interface 
// =====================================

class kr_skm : public skm_algo {

protected:

  // internal variables
  unsigned int Q;          // Prime Q
  unsigned int B;          // Base B
  unsigned int BS;         // Base Shift == log2(B)

  unsigned int Bm;
  unsigned int hpat;
  unsigned int htxt;

  skm_patt::iterator pmpi; // Partial match pattern iterator
  astring::iterator pmti;  // Partial match text iterator

public:

  // constructor / destructor
  kr_skm() : skm_algo() {};
  kr_skm(const alpha &a) : skm_algo(a) {};

  ~kr_skm() {};

  // long algorithm name
  virtual 
  const char * name() const { return "Karp Rabin"; };

  // short algorithm name
  virtual 
  const char * id() const { return "kr"; };

  // print preprocesed data 
  virtual 
  ostream & print_pdata(ostream &o) const;

protected:

  virtual 
  void preprocess ();                // preprocess the pattern

  virtual 
  void run_preprocessor();           // run the preprocessor

  virtual 
  astring::size_type match();       // match the pattern

  virtual
  astring::size_type run_matcher(); // run the matcher

  virtual
  void algorithm_step(breakpoint WHERE, int count);
  // algorithm step function
};

// STATIC OBJECTS
extern kr_skm krskm;

#endif
