// skmmfuns.h -- The -*- C++ -*- single keyword match matching functions.
// Copyright (C) 1997 Sergio Talens-Oliag <sto@uv.es>
// This file is part of the C++ Alphabetic Data Management Library.

#ifndef _SKMMFUNS_H_
#define _SKMMFUNS_H_

#include "skm.h"
#include <iostream.h>
#include <vector>

// =====================================================
// SKM FFMF (Find First Match Function): Class Interface
// =====================================================

class skm_ffmf : public skm_mfun {

protected:

  astring::size_type mi;  // match index
  astring::size_type ts;  // text size

public:

  skm_ffmf () {};   
  ~skm_ffmf() {};

  // start 
  virtual void start (skm_algo *a) { 
    ts = (a->input_text())->size(); 
    mi = ts;
  };

  // match using text indices
  virtual bool found(astring::size_type ti) {
    mi = ti;
    return true; // Stop searching 
  };

  // access methods

  virtual int nmatch() { 
    return ( ts == mi ? 0 : 1 );
  }; // return the number of matches

  astring::size_type match() const {
    return mi;
  };  // return the first match  

protected:

  // Ostream output
  virtual ostream & Out(ostream &o) const;
};

// -------------------
// find first wrappers
// -------------------

inline 
skm_ffmf find_first (skm_algo *s, skm_patt &p, astring &t, 
		     astring::size_type to=0) {
  skm_ffmf ffmf;
  s->find(&ffmf, p, t, to);
  return ffmf;
};

inline 
skm_ffmf find_first_in (skm_algo *s, astring& t, 
			astring::size_type to=0) {
  skm_ffmf ffmf;
  s->find_in(&ffmf, t, to);
  return ffmf;
};

// ===================================================
// SKM FAMF (Find All Match Function): Class Interface
// ===================================================

class skm_famf : public skm_mfun {

protected:

  vector<astring::size_type> mv;  // matches vector
  astring::size_type         ts;

public:

  skm_famf () {};   
  ~skm_famf() {};

  // start 
  virtual void start (skm_algo *a) { 
    mv.erase(mv.begin(), mv.end());  // erase previous results
    ts = (a->input_text())->size(); 
  };

  // match using text indices
  virtual bool found(astring::size_type ti) {
    mv.push_back(ti);
    return false; // Continue searching 
  };

  // access methods

  virtual int nmatch() { 
    return mv.size();
  }; // return the number of matches

  astring::size_type match() const {
    if (mv.size() != 0)
      return mv[0];
    else
      return ts;
  };  // return the first match  

  vector<astring::size_type> matches() const {
    return mv;
  };  // return a copy of the matches vector  

protected:

  // Ostream output
  virtual ostream & Out(ostream &o) const;
};

// -----------------
// find all wrappers
// -----------------

inline 
skm_famf find_all (skm_algo *s, skm_patt &p, astring &t, 
		   astring::size_type to=0) {
  skm_famf famf;
  s->find(&famf, p, t, to);
  return famf;
};

inline 
skm_famf find_all_in (skm_algo *s, astring &t, 
		      astring::size_type to=0) {
  skm_famf famf;
  s->find_in(&famf, t, to);
  return famf;
};


#include "atoken.h"

// ============================================================
// SKM FFTMF (Find First Token Match Function): Class Interface
// ============================================================

class skm_fftmf : public skm_mfun {

protected:

  atoken              *tok;
  skm_patt::size_type ps;   // patt size
  astring             *txt; // text pointer
  astring::size_type  mi;   // match index

public:

  skm_fftmf (atoken *tk) : tok(tk) {};   
  ~skm_fftmf() {};

  // start 
  virtual void start (skm_algo *a) { 
    ps  = (a->input_patt())->size();
    txt = (astring *) a->input_text();
    mi  = txt->size();
  };

  // match using text indices
  virtual bool found(astring::size_type ti) {
    if ( (ti == 0 || tok->is_predelim((*txt)[ti-1]) ) &&
	  tok->is_postdelim((*txt)[ti + ps]) &&
	  tok->is(*txt, ti, ti + ps)){
      mi = ti;
      return true; // Token Found
    } else {
      return false; // Token Not Found, continue
    }
  };

  // access methods

  virtual int nmatch() { 
    return ( mi == txt->size() ? 0 : 1 );
  }; // return the number of matches

  astring::size_type match() const {
    return mi;
  };  // return the match or txt->size()

protected:

  // Ostream output
  virtual ostream & Out(ostream &o) const;
};

// ------------------------
// find first token wrapper
// ------------------------

inline 
skm_fftmf find_token (atoken *tk, skm_algo *s, skm_patt &p,
		      astring &t, astring::size_type to=0) {
  skm_fftmf fftmf(tk);
  s->find(&fftmf, p, t, to);
  return fftmf;
};

inline 
skm_fftmf find_token_in (atoken *tk, skm_algo *s, astring &t, 
			 astring::size_type to=0) {
  skm_fftmf fftmf(tk);
  s->find_in(&fftmf, t, to);
  return fftmf;
};

// ==========================================================
// SKM FATMF (Find All Token Match Function): Class Interface
// ==========================================================

class skm_fatmf : public skm_famf {

protected:

  atoken                     *tok;
  skm_patt::size_type        ps;   // patt size
  astring                    *txt; // text pointer
  vector<astring::size_type> mv;   // matches vector

public:

  skm_fatmf (atoken *tk) : tok(tk) {};   
  ~skm_fatmf() {};

  // start 
  virtual void start (skm_algo *a) { 
    mv.erase(mv.begin(), mv.end());  // erase previous results
    txt = (astring *) a->input_text(); 
    ps  = (a->input_patt())->size();
  };

  // match using text indices
  virtual bool found(astring::size_type ti) {
    if ( (ti == 0 || tok->is_predelim((*txt)[ti-1]) ) &&
	  tok->is_postdelim((*txt)[ti + ps]) &&
	  tok->is(*txt, ti, ti + ps)){
      mv.push_back(ti);
    }
    return false; // Continue Searching
  };

  // access methods

  virtual int nmatch() { 
    return mv.size();
  }; // return the number of matches

  astring::size_type match() const {
    if (mv.size() != 0)
      return mv[0];
    else
      return txt->size();
  };  // return the first match

  vector<astring::size_type> matches() const {
    return mv;
  };  // return a copy of the matches vector  

protected:

  // Ostream output
  virtual ostream & Out(ostream &o) const;
};

// ------------------------
// find all tokens wrappers
// ------------------------

inline 
skm_fatmf find_all_tokens (atoken *tk, skm_algo *s, skm_patt &p, astring &t,
                           astring::size_type to=0) {
  skm_fatmf fatmf(tk);
  s->find(&fatmf, p, t, to);
  return fatmf;
};

inline 
skm_fatmf find_all_tokens_in (atoken *tk, skm_algo *s, astring &t, 
			      astring::size_type to=0) {
  skm_fatmf fatmf(tk);
  s->find_in(&fatmf, t, to);
  return fatmf;
};

#endif
