// skmadev.h -- The -*- C++ -*- single keyword match analisys device interface
// Copyright (C) 1997 Sergio Talens-Oliag <sto@uv.es>
// This file is part of the C++ Alphabetic Data Management Library.

#ifndef _SKMADEV_H_
#define _SKMADEV_H_

#include <string>
#include "skmtypes.h"

/**

  DESCRIPTION

  Base class for the analysis devices.

  This class defines a set of print functions to be implemented by each
  derived class:

    1. Print mpatt (matched pattern)
    2. Print mtext (pattern over text)
    3. Print pdata (always takes a sting as the name of the parameter): 

       - int
       - astring::size_type, astring::difference_type
       - string (generic, transform type in string)

       - patt_vectors : strings, astring::size_type, astring::difference_type
       - alpha_vector : strings, astring::size_type, astring::difference_type

       Note: 

         The size_type values represent positive offsets to move forward in a 
       string and the 'difference_type' represent offsets to move forward or
       backward.

    4. Print message (string)

  Each one has an implicit print area (text_area, pdata_area, messg_area,
  input_area), used when the devices have windows.

  Print action (determined by the values of 'ms' and 'gchf' :

    1. none -> Print and exit
    2. wait -> Print and wait delay miliseconds
    3. getchar_function -> Print string and wait for a char from the user.
                           Return a bool telling if we've to stop.
  
  Input parameters:

    * delay
    * get_char_function

**/

// ===================================================================
// ADEV GETCHFUN (Analysis Device Get Char Function) : Class Interface
// ===================================================================

class adev_getchfun {
protected:
  string s;

public:
  adev_getchfun(const string & msg) : s(msg) {};
  ~adev_getchfun() {};
  
  const string &getmsg() { return s; };

  virtual
  bool operator()(char c) {
    if (c == 'q' || c == 'Q') return true;
    else                      return false;
  };
};

// ==============================================================
// ADEV MPATT (Analysis Device Matched Pattern) : Class Interface
// ==============================================================

class adev_mpatt {
protected:

  astring            *pat;         // pattern
  astring::size_type current_pos;  // current position in pattern 
  bool               have_matches; // Have we matched any symbol?
  vector<bool>       msv;          // Matched Symbols Vector 

public:

  // constructor / destructor
  inline
  adev_mpatt(astring *p = NULL, astring::size_type cp = 0);

  inline
  ~adev_mpatt();

protected:

  // PROTECTED UTILITIES
  inline
  void init_matched ();

  inline
  void reset_matched();

public:

  // INITIAL CONTEXT
  inline
  void init (astring *p, astring::size_type cp);
  
  inline
  void update_pos(astring::size_type ps);

  inline
  void matched_symbols(astring::size_type bidx, astring::size_type eidx);

  // ACCESS METHODS
  inline
  bool matched (astring::size_type i) const;

  inline
  bool current (astring::size_type i) const;

  inline
  astring *pattern() const;

};

// ============================================
// SKM ADEV (Analysis Device) : Class Interface
// ============================================

class skm_adev {

public:

  // typedefs
  typedef skm_alpha      alpha;
  typedef skm_breakpoint breakpoint;
  typedef adev_getchfun  getchfun;

protected:

  int      *  msv;  // Delay after each operation
  getchfun ** gchv; // Get char functions for each operation

public:

  // ========================
  // constructor / destructor
  // ========================

  skm_adev() {
    msv  = new int        [skm_bpc] (0);
    gchv = new getchfun * [skm_bpc] (NULL);
  };

  ~skm_adev(){
    delete []msv;
    delete []gchv;
  };

  // =================
  // Adjust parameters
  // =================

  // check values
  const getchfun * get_getchfun (breakpoint i) const {
    return (skm_bpc > i) ? gchv[i] : (getchfun *)NULL;
  }; 

  const int get_delay(breakpoint i) const {
    return (skm_bpc > i) ? msv[i] : 0;
  }; 

  // update/modify values

  getchfun * set_getchfun (breakpoint i, getchfun * gf = NULL) { 
    if (skm_bpc > i) { getchfun *ogf = gchv[i]; gchv[i] = gf; return ogf; } 
    else             { return NULL; }
  };

  int set_delay (breakpoint i, int dl = 0) { 
    if (skm_bpc > i) { int odl = msv[i]; msv[i] = dl; return odl; } 
    else             { return 0; }
  };

  // =====
  // MATCH
  // =====

  virtual
  void matched_patt (const adev_mpatt &mp) = 0;

  virtual
  void matched_text (astring *txt, astring::iterator ppit,
                     const adev_mpatt &mp) = 0;

  // =====
  // PDATA
  // =====

  // Single values
  virtual
  void int_pdata (const string & name, int value) = 0;

  virtual
  void size_pdata (const string & name, astring::size_type value) = 0;

  virtual
  void diff_pdata (const string & name, astring::difference_type value) = 0;

  virtual
  void str_pdata (const string & name, const string & value) = 0;

  // Patt Vectors
  virtual
  void patt_size_pdata(const string & name, skm_patt * pat,
                       const vector<astring::size_type> &v) = 0;
  virtual
  void patt_diff_pdata(const string & name, skm_patt * pat,
                       const vector<astring::difference_type> &v) = 0;
  virtual
  void patt_str_pdata(const string & name, skm_patt * pat,
                       const vector<string> &v) = 0;

  // Alpha Vectors
  virtual
  void alpha_size_pdata(const string & name, alpha * al,
		       const vector<astring::size_type> &v,
                       astring::size_type def) = 0;

  virtual
  void alpha_diff_pdata(const string & name, alpha * al,
		       const vector<astring::difference_type> &v,
                       astring::difference_type def) = 0;

  virtual
  void alpha_str_pdata(const string & name, alpha * al,
		       const vector<string> &v, string def) = 0;

  // ========
  // Messages
  // ========

  virtual
  void message(const string & msg) = 0; // Print a message 

  // ================
  // User interaction
  // ================
  virtual
  bool user_io(breakpoint i) = 0; // continue running ?

};

// =============================================================
// ADEV MPATT (Analysis Device Matched Pattern) : Inline Methods
// =============================================================

// constructor / destructor
inline
adev_mpatt::adev_mpatt(astring *p, astring::size_type cp)
    : pat(p), current_pos(cp) { 

  if (pat != NULL) init_matched();
  else             have_matches = false;

};
  
inline
adev_mpatt::~adev_mpatt(){};

// PROTECTED UTILITIES

void adev_mpatt::init_matched () {
  msv.erase(msv.begin(), msv.end());
  msv.reserve(pat->size() + 1);
  msv.insert(msv.begin(), pat->size() + 1, false);
  have_matches = false;
};

void adev_mpatt::reset_matched() {
  for (int i=0; i< msv.size(); i++)
    msv[i]=false;
  have_matches = false;
};

// INITIAL CONTEXT
void adev_mpatt::init (astring *p, astring::size_type cp) {
  pat=p;
  current_pos = cp;

  init_matched();
};
  
void adev_mpatt::update_pos( astring::size_type ps){

  current_pos = ps;

  if (have_matches)
    reset_matched();
};

void adev_mpatt::matched_symbols (skm_patt::size_type bidx,
                                  skm_patt::size_type eidx){
  for (int i = bidx; i < eidx; i++)
    msv[i]=true;
  have_matches=true;
};

//
// ACCESS METHODS
//

bool adev_mpatt::matched (skm_patt::size_type i) const {
  return msv[i];
};

bool adev_mpatt::current (skm_patt::size_type i) const {
  return (current_pos == i);
};

skm_patt * adev_mpatt::pattern() const {
  return pat;
};

#endif
