// skmalgo.h -- The -*- C++ -*- single keyword match algorithms interface.
// Copyright (C) 1997 Sergio Talens-Oliag <sto@uv.es>
// This file is part of the C++ Alphabetic Data Management Library.

#ifndef _SKMALGO_H_
#define _SKMALGO_H_

/**

  DESCRIPTION
  
  This header defines the base single keyword search algorithm. 
  
**/

#include "admalgo.h"
#include "skmtypes.h"
#include "skmadev.h"
#include "skmmfun.h"
#include "stepacct.h"

// ==========================
// SKM ALGO : Class Interface
// ==========================

class skm_algo : public adm_algo {

public:

  // typedefs
  typedef skm_alpha       alpha;

  typedef skm_mfun        mfun;

  typedef skm_adev        adev;
  typedef stepacct        stac;

  typedef skm_breakpoint  breakpoint;

protected:

  bool                  stop;    // should we stop running ? used w./ adev

protected:

  // Input data ...

  // ... alphabet support
  alpha                al;        // alphabet
  alpha::size_type     alpha_size; // alphabet_size

  // ... methods for modifying the algorithm behaviour
  mfun                 *mf;        // match function

  // ... for the analysys mode ...

  // ... accounting:
  stac                 sa;         // step accounting
  bool                 use_sa;     // flag to activate/deactivate the step acct 

  // ... displaying:
  adev                 *ad;        // Analysis device

  // ... for each algorithm execution
  skm_patt             *pat;    // pattern
  astring              *txt;    // text
  astring::size_type   txt_off; // initial text offset

  // Internal data for each matching cycle
  skm_patt::iterator   pat_beg;  // patt iterator begin
  skm_patt::iterator   pat_end;  // patt iterator end
  skm_patt::size_type  pat_size; // patt size

  astring::iterator    txt_beg;  // text iterator begin
  astring::iterator    txt_end;  // text iterator end
  astring::size_type   txt_size; // text size

  astring::iterator    ti;       // current text position

public:

  // Constructor / destructor
  inline
  skm_algo (const alpha &a = alpha());

  inline
  ~skm_algo ();

protected:

  //
  // Internal Utilities
  //

  // Alphabet support
  inline
  alpha::size_type aindex(const astring::value_type &s) const;

  // Patt and Text Utilities
  inline
  void update_patt();

  inline
  void set_patt(skm_patt &p);

  inline
  void update_text();

  inline
  void set_text(astring &t);

  // Matcher utilities
  inline
  bool report_match(astring::size_type ti); 

  // Stepper utilities
  inline
  void report_step (breakpoint WHERE, int count = 1);
    // Update step count and exec stepper

public:
  
  //
  // Access to ...
  //

  // ... ALPHABET
  inline
  const alpha use_alpha (const alpha &a = alpha());

  // ... NAMES
  inline
  const char *kind() const; // kind of algorithm

  // ... INPUT DATA
  inline
  const skm_alpha &input_alpha() const; // Return a reference to the alphabet

  inline
  const skm_patt * input_patt() const;   // Return a const pointer to pattern
  
  inline
  const astring  * input_text() const;   // Return a const pointer to text

  inline
  astring::size_type input_text_offset() const; // Return initial text offset

  // ... ACCOUNTING
  inline
  bool accounting() const;              // are we accounting?

  inline
  bool accounting(bool ua);             // turn ON/OFF accounting

  inline
  void reset_accounting ();             // reset step counters

  inline
  const stepacct & steps () const;      // return sa
 
  inline
  int steps (breakpoint i) const;      // return steps on i
  
  // ... DISPLAYING
  inline 
  skm_adev * use_adev(adev *a=NULL);    // Set adev and return the old one

  //
  // Algorithmic methods
  //

  // Preprocess pattern
  void prep(skm_patt &p);
  
  // Find ...
  // ... preprocessed pattern
  astring::size_type find_in(astring &t, astring::size_type to =0);

  // ... pattern in text
  //  inline
  astring::size_type find(skm_patt &p, astring &t, astring::size_type to =0);
  // Explicit invocation

  //  inline
  astring::size_type operator()(skm_patt &p, astring &t,
				astring::size_type to =0);
  // Using the object as a function

  // ... preprocessed pattern using the match function passed
  //  inline
  void find_in(skm_mfun *m, astring &t, astring::size_type to =0);

  // ... pattern in text using the match function passed
  //  inline
  void find(skm_mfun *m, skm_patt &p, astring &t, astring::size_type to =0);
  // Explicit invocation

  //  inline
  void operator()(skm_mfun *m, skm_patt &p, astring &t, 
		  astring::size_type to =0);
  // Using the object as a function

public:

  //
  // PUBLIC PURE VIRTUAL FUNCTIONS
  //

  virtual 
  const char * name() const =0;  // long name of the algorithm
 
  virtual
  const char * id() const =0;    // short name of the algorithm

  virtual 
  ostream & print_pdata(ostream &o) const =0; // print preprocesed data

protected:  

  //
  // PROTECTED PURE VIRTUAL FUNCTIONS
  //
  
  virtual 
  void preprocess()                 =0; // preprocess the pattern

  virtual
  void run_preprocessor()           =0; // run the preprocessor

  virtual 
  astring::size_type match()       =0; // match the pattern

  virtual 
  astring::size_type run_matcher() =0; // run the matcher

  virtual 
  void algorithm_step(breakpoint WHERE, int count) =0;
  // algorithm step function

};

// =========================
// SKM ALGO : Inline Methods
// =========================

// Constructor / destructor
inline
skm_algo::skm_algo (const alpha &a)
  : al(a),
    mf(NULL), 
    ad(NULL), sa(skm_bpnv, skm_bpc), use_sa(false),
    pat(NULL), txt(NULL), txt_off(0) {
  alpha_size = al.size() < al.max_size() ? al.size() + 1 : al.max_size();
};

inline
skm_algo::~skm_algo () {};

//
// Internal Utilities
//

// Alphabet support
inline
const skm_algo::alpha skm_algo::use_alpha (const alpha &a) {
  alpha o (al);
  al = a;
  alpha_size = al.size() < al.max_size() ? al.size() + 1 : al.max_size();
  pat=NULL; // Invalidate preprocessing
  return o;
};

inline
skm_algo::alpha::size_type skm_algo::aindex (const astring::value_type &s) 
  const {
  return al.index(s);
};

// Patt and Text Utilities

// update patt
inline 
void skm_algo::update_patt() {
  pat_beg  = pat->begin(); 
  pat_end  = pat->end();
  pat_size = pat->size();
};

// set patt
inline 
void skm_algo::set_patt(skm_patt &p) {
  pat = &p;
  update_patt();
};

// update text
inline 
void skm_algo::update_text() {
  txt_beg  = txt->begin();
  txt_end  = txt->end();
  txt_size = txt->size();
};

// set text
inline 
void skm_algo::set_text(astring &t) {
  txt = &t;
  update_text();
};

// Matcher utilities

// Report match using text indices
inline 
bool skm_algo::report_match(astring::size_type ti) {
  if (mf == NULL)         // if there's no match function ...
    return true;          // ... stop searching
  else {                  // else ...
    return mf->found(ti); // ... execute match function and return its
                          // value
  }
};

// Stepper utilities

// Update the step count and execute the stepper
inline
void skm_algo::report_step (breakpoint WHERE, int count) {
  if (use_sa) { 
    sa.update(WHERE, count); stop = false;
  }
  if (ad && count) {
    algorithm_step(WHERE, count);
    stop = ad->user_io(WHERE);
  }
};
  
//
// Access to ...
//

// ... NAMES

// Kind of algorithm
inline 
const char * skm_algo::kind() const {
  return "skm"; 
};
// ... INPUT DATA

// Return a const pointer to the alphabet
inline 
const skm_alpha & skm_algo::input_alpha() const {
  return al;
};

// Return a const pointer to the pattern
inline 
const skm_patt * skm_algo::input_patt() const { 
  return pat; 
};

// Return a const pointer to the text
inline 
const astring * skm_algo::input_text() const {
  return txt;
};

// Return a copy of the initial text offset
inline 
astring::size_type skm_algo::input_text_offset() const {
  return txt_off; 
};

// ... ACCOUNTING

// are we accounting?
inline 
bool skm_algo::accounting() const {
  return use_sa;
};

// turn ON/OFF accounting
inline 
bool skm_algo::accounting(bool ua) { 
  bool osa= use_sa; use_sa =ua;
  return osa; 
};

// Reset accounting
inline 
void skm_algo::reset_accounting() { 
  sa.reset();
};

// Return step counters
inline 
const stepacct & skm_algo::steps() const {
  return sa;
};

// number of steps through the breakpoint
inline
int skm_algo::steps(breakpoint i) const {
  return sa.nsteps(i);
};

// ... DISPLAYING

inline 
skm_adev * skm_algo::use_adev(adev *a){
  adev *oad = ad; ad= a;
  return oad;
};

//
// Algorithmic methods
//

///////////////////////
// Find pattern in text

// Explicit invocation
inline
astring::size_type skm_algo::find(skm_patt &p, astring &t,
				  astring::size_type to) {
  prep(p);        // preprocess the pattern ...
  find_in(t, to); // ... and find it in the text
};

// Using the object as a function
inline
astring::size_type skm_algo::operator()(skm_patt &p, astring &t,
					astring::size_type to){
  return find(p,t,to);
};

////////////////////////////////////////////////////////////
// Find preprocessed pattern using the match function passed

inline
void skm_algo::find_in(skm_mfun *m, astring &t, astring::size_type to) {
  mf = m;                // set the match function 
  find_in(t, to);
  mf = NULL;             // forget the match function
}

///////////////////////////////////////////////
// Find pattern using the match function passed

// Explicit invocation

inline
void skm_algo::find(skm_mfun *m, skm_patt &p, astring &t,
		    astring::size_type to) {
  prep(p);               // preprocess the pattern ...
  find_in(m, t, to);     // ... and find it in the text
};

// Using the object as a function
inline
void skm_algo::operator()(skm_mfun *m, skm_patt &p, astring &t,
			  astring::size_type to) {
  find(m, p, t, to);
};

#endif
