// skmmfun.h -- The -*- C++ -*- single keyword match match function interface.
// Copyright (C) 1997 Sergio Talens-Oliag <sto@uv.es>
// This file is part of the C++ Alphabetic Data Management Library.

#ifndef _SKMMFUN_H_
#define _SKMMFUN_H_

/**

  DESCRIPTION
  
  This class is used to define what should be done with a match. 

  Before each single keyword matching run the start method is called
  to give the function a context (skm::idata), no checks are done to
  verify it has been called, our skm class should be derived from
  skm_algo. The start method returns a boolean value to tell the
  matcher if our function will change the text.

  Input parameters on each match:

    ti = match offset (index).
  
  Output parameters on each match: 

    Boolean value telling if we have to continue.

  NOTES ABOUT EDITING IN THE TEXT: 
    
    In a previous version of this class changing the text was allowed,
    but this capability has been removed because it forced all the
    algorithms to restart after a match, and this was already
    supported by the reuse of the preprocessing (with the 'find_in'
    method) and the text offset parameter in the find methods.

    In the old version a modified value of 'ti' was returned after
    each match, but it caused some problems. We'll see now the old
    proposed rules for the value of the updated 'ti' with examples and
    their problems. The examples are for the brute force algorithms,
    the text is 'bababoom', the pattern is 'bab' and the new pattern
    is 'ex', the letter between '[' and ']' indicate the current
    instpected text symbol and the letters between '[*' and '*]'
    indicate the matched part of the text.

    1. if text is not modified return 'ti'. 

       Example:

         found '[*bab*]aboom', continuing the match:
         ... '[*bab*]aboom'   'b[a]boom'       'ba[b]aboom'   ...
	 ... 'ba[*b*]aboom'   'ba[*b*][a]boom  'ba[*ba*]boom' ...
         ... 'ba[*ba*][b]oom' 'ba[*bab*]oom'                  ...
	 found 'ba[*bab*]oom'.

       Problems:

         none

    2. if the pattern is removed return 'ti'

       Example:

         found '[*bab*]aboom', after deletion '[a]bbom'.

       Problems:

         The second 'bab' is not erased -> it's first 'b' is erased
         before the match.

    3. if pattern is replaced return 'ti + new_pat.size() - 1'

       Example: 

         found '[*bab*]aboom', after replace 'ex[a]boom'.

       Problems:

	 if new_pat were 'bab' the result is 'bababoom', the 'ba[*bab*]oom'
         match is lost: [*bab*]aboom -> replace -> bab[a]boom


    4. if new_pat is added as prefix return 'ti + new_pat.size()'

       Example:

         found '[*bab*]aboom', after insertion 'exb[a]baboom' ...
         ... found 'exba[*bab*]oom', after insertion 'exbaexb[a]boom'

       Problems:

         none

    5. if new_pat is added as suffix, return 'ti + pat.size() +
       new_pat.size() - 1'

       Example:

         found '[*bab*]aboom', after insertion 'babex[a]boom' 

       Problems:

	 the second 'bab' is not found.

**/

#include <iostream.h>
#include "skmtypes.h"

// ===========================================
// SKM MFUN (Match Function) : Class Interface
// ===========================================

class skm_mfun {

public:

  // constructor / destructor
  skm_mfun(){};
  ~skm_mfun(){};
  
  // start 
  virtual void start (skm_algo *a) {};

  // match using text indices
  virtual bool found(astring::size_type ti) { 
    return false; // Continue searching
  };

  // access methods
  virtual int nmatch() { return 0; }; // return the number of matches

  // Ostream output
  virtual ostream & Out(ostream &o) const { return o; };
  friend  ostream & operator<< (ostream &o, const skm_mfun &mf);

};

// ==========================================
// SKM MFUN (Match Function) : Inline Methods
// ==========================================

// match function output
inline
ostream & operator<< (ostream &o, const skm_mfun &mf) {
  return mf.Out(o);
};

#endif
