/*  cskmtest.cc -- Test program for the -*- C++ -*- skm algorithms (curses)
    Copyright (C) 1997 Sergio Talens-Oliag <sto@uv.es>

    This file is part of the C++ Alphabetic Data Management Library. */

#include "skmtest.h"
#include "stopw.h"

#include "skmcadev.h"
#include "util.h"     // ltos()

#include <getopt.h>
#include <stdlib.h>  // strtoul()

// Program options and usage
static const struct option long_options[]= {
  { _("algorithm"), required_argument, NULL,'a' },  // Select algorithm to use
  { _("alphabet"), required_argument, NULL,'A' },   // Use given alphabet
  { _("breakpoints"), required_argument, NULL,'b'}, // Stop points
  { _("delays"), required_argument, NULL,'d' },     // Delays
  { _("delay-time"), optional_argument, NULL,'D' }, // Delay Time in ms.
  { _("find-first"), no_argument, NULL,'f' },       // Find first occurrence
  { _("help"), no_argument, NULL, 'h' },            // Print usage
  { _("no-prep"), no_argument, NULL, 'n' },         // No preprocessor
  { _("patterns"), required_argument, NULL,'p' },   // Patterns File
  { _("time"), no_argument, NULL, 't' },            // Time algorithms
  { _("text-size"), required_argument, NULL, 'T' }, // Simulated Text Size
  { _("version"), no_argument, NULL, 'v' },         // Show Version
  { 0, 0, 0, 0 }
};

// Usage functions
void short_usage   (int status);
void usage         (int status);

// Data output
void print_wline    (NCursesWindow *w);
void print_input    (NCursesWindow *w, bool custom_alpha, int nor);
void print_algoname (NCursesWindow *w, skm_algo *a); 
void print_results  (NCursesWindow *w, skm_algo *a, skm_mfun *mf, 
		     bool timed, double laptime);

// ===============
// USAGE FUNCTIONS
// ===============

void short_usage (int status) {
  show_version();

  cout << _("Usage: ") << program_name << _(" [OPTIONS]... [PATTERN] FILE\n");
  help_msg(status);
};

void usage (int status) {
  show_version();

  cout << _("Usage: ") << program_name << _(" [OPTIONS]... [PATTERN] FILE\n");
  cout << _("\n\
The options are:\n\
\n\
  -a, --algorithm[=NAME] select an algorithm to use (use all if not present)\n\
  -A, --alphabet[=CHARS] use given alphabet\n\
  -b, --breakpoints[=BP] stop at given breakpoints\n\
  -d, --delay[=BP]       wait on given breakpoints\n\
  -D, --delay-time[=MS]  time to wait on the breakpoints (in miliseconds)\n\
  -f, --find-first       find only the first occurrence of the pattern\n\
  -h, --help             this help\n\
  -n, --no-prep          supress preprocessing\n\
  -p, --patterns[=FILE]  read patterns from FILE (don't pass one on the\n\
                         command line)\n\
  -t, --time             time algorithms\n\
  -T, --text-size[=SIZE] run each algorithm multiple times to simulate\n\
                         a text of SIZE bytes (only used with -t)\n\
  -v, --version          version information\n\
\n\
Where:\n\
\n\
  BP     comma separated list of breakpoints (leave blank to see list)\n\
  CHARS  set of chars on the alphabet (string)\n\
  FILE   file name, each line of the patterns file will be treated as one\n\
         pattern\n\
  NAME   algorithm to use (leave it blank to see acceptable values,\n\
         put multiple -a options to use more than one)\n\
  SIZE   integer value telling the size to simulate, a suffix can be used\n\
         to tell the units: 'b' for bytes, 'k' for Kbytes and 'm' for \n\
         Mbytes (the default is 'b')\n\
\n\
Note:\n\
  The following C escape symbols can be used in CHARS or PATTERN:\n\
      \\n    \\t    \\v    \\b    \\r    \\f    \\a\n\
      \\\\    \\?    \\\'    \\\"    \\ooo  \\xhh\n\
  but see your shell manual to see how to quote (in bash use $\'STRING\')\n\
\n");
  cout << endl;
  exit (status);
};

// ===========
// Data Output
// ===========

void print_wline(NCursesWindow *w){
  string s(w->width(), ' ');
  int y, x;
  w->getyx(y, x);
  w->addnstr(y, 0, s.begin(), s.size());
}

void print_algoname (NCursesWindow *w, skm_algo *a) {
  string s;
  s = "-*- "; s += a->name(); s += " -*-";
  w->erase();
  print_wline(w);
  w->move(0, (w->width() - s.size()) / 2);
  w->addnstr(s.begin(), s.size());
  w->refresh();
}

void print_input (NCursesWindow *w, bool custom_alpha, int nor) {
  string s;

  w->erase();

  s  = _("INPUT DATA");
  w->addnstr (s.begin(), s.size());
  w->addch('\n');

  s  = _("text       : ");
  s += txt_filename;
  s += " ("; s += ltos(txt.size()); s +=")";
  w->addnstr (s.begin(), s.size());
  w->addch('\n');

  s  = _("pattern    : ");
  s += pat;
  s += " ("; s += ltos (pat.size()); s += ")";
  w->addnstr (s.begin(), s.size());
  w->addch('\n');

  if (custom_alpha) {
    s  = _("using custom alpha");
    s += " ("; s += ltos (al.size()); s += ")";
    w->addnstr (s.begin(), s.size());
    w->addch('\n');
  }

  if (nor != 1) {
    s  = _("multiple runs");
    s += " ("; s += ltos (nor); s += ")";
    s += _("\tsimulated text size");
    s += " ("; s += ltos (nor * txt.size()); s += ")";
    w->addnstr (s.begin(), s.size());
    w->addch('\n');
  }

  w->refresh();
};

void print_results (NCursesWindow *w, skm_algo *a, skm_mfun *mf, 
		    bool timed, double laptime) {
  string         m;
  const stepacct sa = a->steps();
  int            i; 

  w->erase();

  // Print Steps (includes number of matches)
  i = 0;
  while (i < sa.nbps()) {
    m = sa.bpname(i); m += " : "; m += sa.nsteps(i); i++; 
    if (i % 2 != 0) 
      m += '\t'; 
    else
      m += '\n'; 
  }

  w->addnstr(m.begin(),  m.size());

  if (timed) {
    m  = _("elapsed time       : ");
    m += ftos(laptime / timer.resolution());
    w->addnstr(m.begin(),  m.size());
  }

  w->refresh();
};

// ============
// Main program
// ============

int main (int argc, char *argv[]) {

  // SELECT ALGORITHMS
  vector<bool>        skmr;          // Run algorithm i iff skmr[i] == true

  // MATCH FUNCTIONS
  skm_mfun            *matcher;
  skm_ffmf            ffmf;
  skm_famf            famf;

  // ANALYSIS DEVICES
  skm_cadev           *cursadev;

  // Program options
  bool                all;           // All algorithms selected?
  bool                custom_alpha;  // Are we using a custom alphabet ?
  bool                find_first;    // Find only first occurrence?
  bool                noprep;        // Don't preprocess?
  bool                timed;         // Run StopWatch ?

  // Algorithms timming
  double              laptime;       // Elapsed time

  // Simulate a bigger text
  unsigned int        stsz; // Simulated Text Size
  unsigned int        nor;  // Number of runs to simulate size

  // auxiliary variables
  int  i, optchar;
  char *s;

  // get char function 
  adev_getchfun ag(_("Pulse \'q\' para SALIR u otra tecla para CONTINUAR"));
  int           dt(1000); // delay time = 1 second

  // delay / get char vectors
  bool dlv [skm_bpc]; 
  bool gcv [skm_bpc];

  // ========
  // Init NLS
  // ========

  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);

  // ==================
  // Set default values
  // ==================

  custom_alpha = false;

  all          = true;
  noprep       = false;
  find_first   = false;
  timed        = false;

  stsz         = 0;
  nor          = 1;

  // Search algorithms
  skmv.push_back(&bfskm);   // Brute Force
  skmv.push_back(&bmhskm);  // Boyer Moore Horspool
  skmv.push_back(&bmskm);   // Boyer Moore
  skmv.push_back(&kmpskm);  // Knuth Morris Pratt
  skmv.push_back(&krskm);   // Karp Rabin
  skmv.push_back(&soskm);   // Shift Or
  skmv.push_back(&sqsskm);  // Sunday Quick Search

  // Don't execute algorithms
  skmr.insert(skmr.begin(), skmv.size(), false); 

  // delay / get char
  for (i=0; i<skm_bpc; i++) {
    dlv[i] = true; gcv[i] = false;
  }

  // =============
  // Parse Options
  // =============

  // Read Program Name
  program_name = argv[0];

  opterr = 0; // Don't print default errors

  while ((optchar = getopt_long(argc, argv, "a:A:b:d:D:fhnp:stT:v",
				long_options, NULL)) != EOF) {

    switch (optchar) {
      
    case 0:
      break;

    case '?':
      switch (optopt) {
      case 'a': show_algonames();
                break;
      case 'A': cerr << _("Give an alphabet") << endl; help_msg(1);
                break;
      case 'b':
      case 'd': show_breakpoints();
                break;
      case 'p': cerr << _("Give a patterns file name") << endl; help_msg(1);
                break;
      case 'T': cerr << _("Give a simulated text size") << endl; help_msg(1);
                break;
      default : cerr << char(optopt) << _(": Unknown option") << endl;
                help_msg(1);
                break;
      }
      break;

    case 'a':                           // SELECT ALGORITHMS
      for (i=0; i<skmv.size(); i++) {
        if (! strcmp(optarg, skmv[i]->id()))
          skmr[i]= true; all= false;
      }
      // Error selecting algorithm
      if (all)
	show_algonames();
      break;

    case 'A':                           // Read an ALPHABET
      if (optarg != NULL) {
        vector<char> va;
        char         *s;

        for (s = optarg; *s; s++)
	  va.push_back(parse_char(s, &s));

        al = skm_alpha(va); // optarg is a char vector

        for (i=0; i < skmv.size(); i++)
          skmv[i]->use_alpha(al);

        custom_alpha = true;
      }
      break;

    case 'b':                           // Breakpoints
    case 'd':                           // Delays
      if (optarg != NULL) {
        char *sb, *sc;

        // reset delay vector
	if (optchar == 'd') 
          for (i=0; i<skm_bpc; i++) dlv[i] = false;

        sb = optarg;
        
	while (*sb != '\0') {
          i = strtoul (sb, &sc, 10);
          if (i < skm_bpc && sb != sc) {
	    if (optchar == 'b')      gcv[i] = true;
	    else if (optchar == 'd') dlv[i] = true;
	  } else {
            cerr << sb << " : Unknown breakpoint" << endl;
            short_usage (1);
          }
	  while (*sc != '\0') {
	    if (isdigit(*sc)) break;
            sc++;
          } 
          sb = sc;
        }
      }
      break;

    case 'D':                           // Delay Time
      dt = strtoul(optarg, &s, 10);
      if (optarg == s) {
        cerr << "Invalid delay" << endl;
        short_usage(1);
      }
      break;

    case 'h':                           // HELP
      usage(0); break;

    case 'f':                           // Find first ocurrence only
      find_first = true; break;

    case 'n':                           // Don't preprocess
      noprep = true; break;

    case 'p':                           // Read patterns from file
      if (!read_pattv (optarg)) {
        cerr << _("Error reading patterns file") << endl;
        help_msg(1);
      }
      break;

    case 't':                           // TIME ALGORITHMS
      timed  = true; break;

    case 'T':                           // Text size
      stsz = strtol(optarg, &s , 10);
      switch (*s) {
      case '\0': case 'B': case 'b':
	break;
      case 'K': case 'k':
	stsz *= 1024; break;
      case 'M': case 'm':
	stsz *= 1024 * 1024; break;
      default:
        cerr << char(*s) << _(": Unknown size unit") << endl;
	help_msg(1); break;
      };
      break;

    case 'v':                           // VERSION
      show_version();
      exit(0);
      break;

    default:
      short_usage (1);

    } // end switch
  } // end while

  // ==============
  // PATTERN & TEXT
  // ==============

  // Read PATTERN
  if (pv.size() == 0) {

    if (argc - optind != 2) // We need 2 more arguments
      short_usage(1);

    if (! read_patt (argv[optind++])) {
      cerr << _("Error parsing the pattern\n") << endl;
      short_usage(1);
    }

  } else {
    if (argc - optind != 1) // We need one more argument
      short_usage(1);    
  }

  // Read TEXT
  txt_filename = argv[optind];
  if (!read_text()) {
    cerr << _("Error opening text file: ") << txt_filename << endl;
    help_msg(1);
  };

  // =================
  // Adjust Parameters
  // =================

  // Number of runs (one if not timed)
  if (stsz != 0 && timed) {
    int ts = txt.size(); 
    if (ts < stsz)
      nor = stsz / ts;
  }

  // Select Matcher
  if (find_first) 
    matcher = &ffmf;
  else
    matcher = &famf;

  // Select all Algorithms
  if (all) {
    // If no algorithm selected, use all
    for (i=0; i < skmv.size(); i++) 
      skmr[i]=true;
  }

  // ============
  // Setup Curses
  // ============

  // Main Window
  NCursesWindow progw (stdscr);

  // Algorithm name (1 line)
  NCursesWindow algow (progw, 1, progw.width(), 0, 0, 'r');

  // Analysis Window (screen lines - 7)
  NCursesWindow analw (progw, progw.height()-8, progw.width(), 1, 0, 'r');

  // Data Window (5 lines)
  NCursesWindow dataw (progw, 5, progw.width(), progw.height()-7, 0, 'r');

  // Title Window (1 line)
  NCursesWindow titlw (progw, 1, progw.width(), progw.height()-2, 0, 'r');
  
  // ===============
  // Start Up CURSES
  // ===============

  NCursesWindow::useColors(); // Use Colors if avaliable
  curs_set(0);                // Hide Cursor

  // ============
  // Title Window
  // ============

  string title (program_name);
  title += " " PACKAGE " (" VERSION ")";

  titlw.attron(A_REVERSE);
  print_wline(&titlw);

  titlw.move(0, (titlw.width() - title.size())/ 2);
  titlw.addnstr(title.begin(), title.size());

  titlw.attroff(A_REVERSE);
 
  // ==============
  // Algorithm Name
  // ==============

  algow.attron(A_REVERSE);
  print_wline(&algow);

  // ===========
  // CURSES ADEV
  // ===========

  cursadev = new skm_cadev (&analw);

  // set delay & getchar
  for (int bi=0; bi < skm_bpc; bi++) {
    if (dlv[bi]) cursadev->set_delay(skm_breakpoint(bi), dt);
    else         cursadev->set_delay(skm_breakpoint(bi), 0);

    if (gcv[bi]) cursadev->set_getchfun(skm_breakpoint(bi), &ag);
    else         cursadev->set_getchfun(skm_breakpoint(bi), NULL);
  }

  // Paint screen
  progw.refresh();

  // Wait delay time ... :-)
  napms(dt);

  // =================
  // PROGRAM EXECUTION
  // =================

  do {

    // --------------------
    // Get the next pattern
    // --------------------

    if (pv.size()!=0) {
      pat = pv.front();
      pv.pop_front();
    }

    // ----------------------------------
    // Preprocess Pattern if not using it
    // ----------------------------------

    if (noprep) {
      for (i=0; i < skmv.size(); i++)
	skmv[i]->prep(pat);
    }

    // -----------------------------------
    // Run each algorithm over the pattern
    // -----------------------------------

    for (i=0; i < skmv.size(); i++) {
      
      if (skmr[i]) {

	// Show algoname
	print_algoname(&algow, skmv[i]);

        // Print input data
        print_input(&dataw, custom_alpha, nor);

        // Clear Analysis Window
        analw.erase(); 
        analw.refresh();

	skmv[i]->use_adev(cursadev);

        skmv[i]->accounting(true);
        skmv[i]->reset_accounting();

        // Run the algorithm
	if (noprep)  skmv[i]->find_in(matcher, txt);
	else         skmv[i]->find(matcher, pat, txt);

        // Don't use adev again
	skmv[i]->use_adev();

        // Turn off accounting
	skmv[i]->accounting(false);

        // If timed, run nor times and get laptime
        if (timed) {
	  timer.start();
	  if (noprep) {
	    for (int t=0; t<nor; t++)
	      skmv[i]->find_in(matcher, txt);
	  } else {
	    for (int t=0; t<nor; t++)
	      skmv[i]->find(matcher, pat, txt);
	  }
          laptime = timer.stop();
        }

        // Print results
	print_results(&dataw, skmv[i], matcher, timed, laptime);

      } // end if
      
    } // end for

  } while (pv.size() != 0);

  // Wait delay time ... :-)
  napms(dt);

  // Curses CleanUP
  curs_set(1);                // Show Cursor
  delete cursadev;            // Delete Analysis Device
};
