/*  skmtest.cc -- Test program for the -*- C++ -*- skm algorithms (text)
    Copyright (C) 1997 Sergio Talens-Oliag <sto@uv.es>

    This file is part of the C++ Alphabetic Data Management Library. */

#include "skmtest.h"
#include "stopw.h"

#ifdef __macos__
# include "mac_main.h"
#endif

#include <getopt.h>

// Program options and usage
static const struct option long_options[]= {
  { _("algorithm"), required_argument, NULL,'a' },  // Select algorithm
  { _("alphabet"), required_argument, NULL,'A' },   // Use given alphabet
  { _("breakpoints"), required_argument, NULL,'b'}, // Stop points
  { _("compact"), no_argument, NULL, 'c' },         // Print compact results
  { _("delays"), required_argument, NULL,'d' },     // Delays
  { _("delay-time"), optional_argument, NULL,'D' }, // Delay Time in ms.
  { _("find-first"), no_argument, NULL,'f' },       // Find first occurrence
  { _("help"), no_argument, NULL, 'h' },            // Print usage
  { _("interactive"), no_argument, NULL, 'i' },     // Run interactive
  { _("no-prep"), no_argument, NULL, 'n' },         // No preprocessor
  { _("patterns"), required_argument, NULL,'p' },   // Patterns File
  { _("report"), no_argument, NULL, 'r' },          // Report Steps
  { _("show"), no_argument, NULL, 's' },            // Show execution
  { _("time"), no_argument, NULL, 't' },            // Time algorithms
  { _("text-size"), required_argument, NULL, 'T' }, // Simulated Text Size
  { _("version"), no_argument, NULL, 'v' },         // Show Version
  { 0, 0, 0, 0 }
};

// Usage functions
void short_usage   (int status);
void usage         (int status);

// Data output
void print_input    (ostream &o, bool custom_alpha, int nor);
void print_algoname (ostream &o, skm_algo *a); 
void print_results  (ostream &o, skm_algo *a, skm_mfun *mf, 
		     bool report, bool timed, double laptime);
void print_cinit    (ostream &o, bool custom_alpha, int nor);
void print_cpatt    (ostream &o, bool custom_alpha, int nor);
void print_cresults (ostream &o, skm_algo *a, skm_mfun *mf, 
		     bool report, bool timed, double laptime);

// ===============
// USAGE FUNCTIONS
// ===============

void short_usage (int status) {
  show_version();

  cout << _("Usage: ") << program_name << _(" [OPTIONS]... [PATTERN] FILE\n");
  help_msg(status);
};

void usage (int status) {
  show_version();

  cout << _("Usage: ") << program_name << _(" [OPTIONS]... [PATTERN] FILE\n");
  cout << _("\n\
The options are:\n\
\n\
  -a, --algorithm[=NAME] select an algorithm to use (use all if not present)\n\
  -A, --alphabet[=CHARS] use given alphabet\n\
  -b, --breakpoints[=BP] stop at given breakpoints\n\
  -c, --compact          uses a compact output format for -r and -t \n\
  -d, --delay[=BP]       wait on given breakpoints\n\
  -D, --delay-time[=MS]  time to wait on the breakpoints (in miliseconds)\n\
  -f, --find-first       find only the first occurrence of the pattern\n\
  -h, --help             this help\n\
  -i, --interactive      run interactive\n\
  -n, --no-prep          supress preprocessing\n\
  -p, --patterns[=FILE]  read patterns from FILE (don't pass one on the\n\
                         command line)\n\
  -r, --report           report algorithm steps\n\
  -s, --show             show algorithm execution\n\
  -t, --time             time algorithms\n\
  -T, --text-size[=SIZE] run each algorithm multiple times to simulate\n\
                         a text of SIZE bytes (only used with -t)\n\
  -v, --version          version information\n\
\n\
Where:\n\
\n\
  BP     comma separated list of breakpoints (leave blank to see list)\n\
  CHARS  set of chars on the alphabet (string)\n\
  FILE   file name, each line of the patterns file will be treated as one\n\
         pattern\n\
  NAME   algorithm to use (leave it blank to see acceptable values,\n\
         put multiple -a options to use more than one)\n\
  SIZE   integer value telling the size to simulate, a suffix can be used\n\
         to tell the units: 'b' for bytes, 'k' for Kbytes and 'm' for \n\
         Mbytes (the default is 'b')\n\
\n\
Notes:\n\
1. When simulating a bigger text (option -T), the matches are only for one\n\
   run (it makes no sense report each position more than once).\n\
2. Options -b, -d and -D only make sense if running with -s and -i.\n\
2. The following C escape symbols can be used in CHARS or PATTERN:\n\
       \\n    \\t    \\v    \\b    \\r    \\f    \\a\n\
       \\\\    \\?    \\\'    \\\"    \\ooo  \\xhh\n\
   but see your shell manual to see how to quote (in bash use $\'STRING\')\n\
\n");
  cout << endl;
  exit (status);
};

// ===========
// Data Output
// ===========

void print_input (ostream &o, bool custom_alpha, int nor) {
  o << _("\n-*- Input Data -*-\n\n");
  o << _("text               : ") << txt_filename << "\n";
  o << _("text size          : ") << txt.size()   << "\n";
  o << _("pattern            : ") << pat          << "\n";
  o << _("pattern size       : ") << pat.size()   << "\n";
  if (custom_alpha) {
    o << _("alphabet           : ") << al         << "\n";
    o << _("alphabet size      : ") << al.size()  << "\n";
  }
  if (nor != 1) {
    o << _("search loops       : ") << nor              << "\n";
    o << _("effective text size: ") << nor * txt.size() << "\n";
  }
  o << endl;
};

void print_algoname (ostream &o, skm_algo *a) {
  o << "-*- " << a->name() << " -*-\n\n";
}

void print_results (ostream &o, skm_algo *a, skm_mfun *mf, 
		    bool report, bool timed, double laptime) {
  print_algoname (o,a);
  o << *mf;
  if (timed)
    o << "\n" << _("elapsed time       : ")
      << (laptime / timer.resolution()) << "\n";
  if (report)
    o << "\n" << a->steps();
  o << endl;
};

void print_cinit (ostream &o, bool custom_alpha, int nor) {
  o << _("--- input ---\n");
  o << _("txt:") << txt.size() << ":" << txt_filename << "\n";
  if (custom_alpha)
    o << _("alf:") << al.size() << ":" << al << "\n";
  if (nor != 1)
    o << _("sts:") << nor << ":" << nor * txt.size() << "\n"; 
  o << _("--- output format ---\n");
  o << _("algo:pat_size:num_match:lap_time:step_counters\n");
};

void print_cpatt (ostream &o) {
  o << "---\n";
  o << _("pat:") << pat.size() << ":" << pat << "\n";
};

void print_cresults (ostream &o, skm_algo *a, skm_mfun *mf, 
	             bool report, bool timed, double laptime) {
  o << a->id();
  o << ":";
  o << pat.size();
  o << ":";
  o << mf->nmatch();
  o << ":";
  if (timed)
    o << (laptime / timer.resolution());
  o << ":";
  if (report) {
    const stepacct sa = a->steps();
    int i = 0;

    o << sa.nsteps(i++); 
    while (i < sa.nbps())
      o << "," << sa.nsteps(i++); 
  }
  o << endl;
};

// ============
// Main program
// ============

int main (int argc, char *argv[]) {

  // SELECT ALGORITHMS
  vector<bool>        skmr;          // Run algorithm i iff skmr[i] == true

  // MATCH FUNCTIONS
  skm_mfun            *matcher;
  skm_ffmf            ffmf;
  skm_famf            famf;

  // ANALYSIS DEVICES
  skm_tadev           textadev;

  // Program options
  bool                all;           // All algorithms selected?
  bool                compact;       // Are we using a compact output
  bool                custom_alpha;  // Are we using a custom alphabet ?
  bool                find_first;    // Find only first occurrence?
  bool                interactive;   // Run interactive ?
  bool                noprep;        // Don't preprocess?
  bool                report;        // Use Accounting ?
  bool                show;          // Show Algorithms Execution ?
  bool                timed;         // Run StopWatch ?

  // Algorithms timming
  double              laptime;       // Elapsed time

  // Simulate a bigger text
  unsigned int        stsz; // Simulated Text Size
  unsigned int        nor;  // Number of runs to simulate size

  // auxiliary variables
  int  i, optchar;
  char *s;

  // get char function for interactive runs
  adev_getchfun ag(_("Introduzca un caracter (\'q\'= SALIR, otro= SEGUIR) "));
  int           dt(1000); // delay time = 1 second

  // delay / get char vectors
  bool dlv [skm_bpc]; 
  bool gcv [skm_bpc];


  // ========
  // Init NLS
  // ========

  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);

  // ==================
  // Set default values
  // ==================


  all          = true;
  compact      = false;
  custom_alpha = false;
  noprep       = false;
  interactive  = false;
  find_first   = false;
  report       = false;
  show         = false;
  timed        = false;

  stsz         = 0;
  nor          = 1;

  // Search algorithms
  skmv.push_back(&bfskm);   // Brute Force
  skmv.push_back(&bmhskm);  // Boyer Moore Horspool
  skmv.push_back(&bmskm);   // Boyer Moore
  skmv.push_back(&kmpskm);  // Knuth Morris Pratt
  skmv.push_back(&krskm);   // Karp Rabin
  skmv.push_back(&soskm);   // Shift Or
  skmv.push_back(&sqsskm);  // Sunday Quick Search

  // Don't execute algorithms
  skmr.insert(skmr.begin(), skmv.size(), false); 

  // =============
  // Parse Options
  // =============

  // Read Program Name
  program_name = argv[0];

  opterr = 0; // Don't print default errors

  while ((optchar = getopt_long(argc, argv, "a:A:b:cd:D:fhinp:rstT:v",
				long_options, NULL)) != EOF) {

    switch (optchar) {
      
    case 0:
      break;

    case '?':
      switch (optopt) {
      case 'a': show_algonames();
                break;
      case 'A': cerr << _("Give an alphabet") << endl; help_msg(1);
                break;
      case 'b':
      case 'd': show_breakpoints();
                break;
      case 'p': cerr << _("Give a patterns file name") << endl; help_msg(1);
                break;
      case 'T': cerr << _("Give a simulated text size") << endl; help_msg(1);
                break;
      default : cerr << char(optopt) << _(": Unknown option") << endl;
                help_msg(1);
                break;
      }
      break;

    case 'a':                           // SELECT ALGORITHMS
      if (optarg != NULL) {
	for (i=0; i<skmv.size(); i++) {
	  if (! strcmp(optarg, skmv[i]->id())) {
	    skmr[i]= true; all= false;
	  }
	}
      }
      // Error selecting algorithm
      if (all)
	show_algonames();
      break;

    case 'A':                           // Read an ALPHABET
      if (optarg != NULL) {
        vector<char> va;
        char         *s;

        for (s = optarg; *s; s++)
	  va.push_back(parse_char(s, &s));

        al = skm_alpha(va); // optarg is a char vector

        for (i=0; i < skmv.size(); i++)
          skmv[i]->use_alpha(al);

        custom_alpha = true;
      }
      break;

    case 'b':                           // Breakpoints
    case 'd':                           // Delays
      if (optarg != NULL) {
        char *sb, *sc;

        // reset delay vector
	if (optchar == 'd') 
          for (i=0; i<skm_bpc; i++) dlv[i] = false;

        sb = optarg;
        
	while (*sb != '\0') {
          i = strtoul (sb, &sc, 10);
          if (i < skm_bpc && sb != sc) {
	    if (optchar == 'b')      gcv[i] = true;
	    else if (optchar == 'd') dlv[i] = true;
	  } else {
            cerr << sb << " : Unknown breakpoint" << endl;
            short_usage (1);
          }
	  while (*sc != '\0') {
	    if (isdigit(*sc)) break;
            sc++;
          } 
          sb = sc;
        }
      }
      break;

    case 'D':                           // Delay Time
      dt = strtoul(optarg, &s, 10);
      if (optarg == s) {
        cerr << "Invalid delay" << endl;
        short_usage(1);
      }
      break;

    case 'c':                           // Compact Output
      compact = true; break;

    case 'f':                           // Find first ocurrence only
      find_first = true; break;

    case 'h':                           // HELP
      usage(0); break;

    case 'i':                           // Run Interactive
      interactive = true; break;

    case 'n':                           // Don't preprocess
      noprep = true; break;

    case 'p':                           // Read patterns from file
      if (optarg != NULL)
	if (!read_pattv (optarg)) {
	  cerr << _("Error reading patterns file") << endl;
	  help_msg(1);
	}
      break;

    case 'r':                           // REPORT STEPS
      report = true; break;

    case 's':                           // SHOW EXECUTION
      show   = true; break;

    case 't':                           // TIME ALGORITHMS
      timed  = true; break;

    case 'T':                           // Text size
      stsz = strtol(optarg, &s , 10);
      switch (*s) {
      case '\0': case 'B': case 'b':
	break;
      case 'K': case 'k':
	stsz *= 1024; break;
      case 'M': case 'm':
	stsz *= 1024 * 1024; break;
      default:
        cerr << char(*s) << _(": Unknown size unit") << endl;
	help_msg(1); break;
      };
      break;

    case 'v':                           // VERSION
      show_version();
      exit(0);
      break;

    default:
      short_usage (1);

    } // end switch
  } // end while

  // ==============
  // PATTERN & TEXT
  // ==============

  // Read PATTERN
  if (pv.size() == 0) {

    if (argc - optind != 2) // We need 2 more arguments
      short_usage(1);

    if (! read_patt (argv[optind++])) {
      cerr << _("Error parsing the pattern\n") << endl;
      short_usage(1);
    }

  } else {
    if (argc - optind != 1) // We need one more argument
      short_usage(1);    
  }

  // Read TEXT
  txt_filename = argv[optind];
  if (!read_text()) {
    cerr << _("Error opening text file: ") << txt_filename << endl;
    help_msg(1);
  };

  // =================
  // Adjust Parameters
  // =================

  // Number of runs (one if not timed or reporting)
  if (stsz != 0 && timed) {
    int ts = txt.size(); 
    if (ts < stsz)
      nor = stsz / ts;
  }

  // Select Matcher
  if (find_first) 
    matcher = &ffmf;
  else
    matcher = &famf;

  // Select all Algorithms
  if (all) {
    // If no algorithm selected, use all
    for (i=0; i < skmv.size(); i++) 
      skmr[i]=true;
  }

  // =================
  // PROGRAM EXECUTION
  // =================

  show_version();

  if (compact)
    print_cinit(cout, custom_alpha, nor);

  do {

    // --------------------
    // Get the next pattern
    // --------------------

    if (pv.size()!=0) {
      pat = pv.front();
      pv.pop_front();
    }

    // ----------------------------------
    // Preprocess Pattern if not using it
    // ----------------------------------

    if (noprep) {
      for (i=0; i < skmv.size(); i++)
	skmv[i]->prep(pat);
    }

    // ----------------
    // Print input data
    // ----------------

    if (compact)
      print_cpatt(cout);
    else
      print_input(cout, custom_alpha, nor);

    // set delay & getchar 
    if (show && interactive) {
      for (int bi=0; bi < skm_bpc; bi++) {
        if (dlv[bi]) textadev.set_delay(skm_breakpoint(bi), dt);
        else         textadev.set_delay(skm_breakpoint(bi), 0);
      
        if (gcv[bi]) textadev.set_getchfun(skm_breakpoint(bi), &ag);
        else         textadev.set_getchfun(skm_breakpoint(bi), NULL);
      }
    }

    // -----------------------------------
    // Run each algorithm over the pattern
    // -----------------------------------

    for (i=0; i < skmv.size(); i++) {
      
      if (skmr[i]) {

        // If showing, set ADEV and print algorithm name
        if (show) {
          skmv[i]->use_adev(&textadev);
          // Show algoname
          print_algoname(cout, skmv[i]);
        }

        // If reporting, turn on accounting
        if (report) {
          skmv[i]->accounting(true);
          skmv[i]->reset_accounting();
        }

        // Run the algorithm
        if ( show || report || ! timed ) {
	  if (noprep)  skmv[i]->find_in(matcher, txt);
	  else         skmv[i]->find(matcher, pat, txt);
	}

        // If we were showing, don't use adev again
	if (show)   skmv[i]->use_adev(NULL);

        // If we were reporting turn off accounting
	if (report) skmv[i]->accounting(false);

        // If timed, run nor times and get laptime
        if (timed) {
	  timer.start();
	  if (noprep) {
	    for (int t=0; t<nor; t++)
	      skmv[i]->find_in(matcher, txt);
	  } else {
	    for (int t=0; t<nor; t++)
	      skmv[i]->find(matcher, pat, txt);
	  }
          laptime = timer.stop();
        }

        // Print results
	if (compact)
	  print_cresults(cout, skmv[i], matcher, report, timed, laptime);
	else
	  print_results(cout, skmv[i], matcher, report, timed, laptime);

      } // end if
      
    } // end for

  } while (pv.size() != 0);

};
